/*
    pool.h - linear tree and collision pool handling (kd-tree-2)
    
    Copyright (C) 2009  Laurent Poirrier
    This file is part of the lin-kd-tree library.

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __KDTREE_POOL_H
#define __KDTREE_POOL_H
#include <stdint.h>
#include "space.h"
#include "kd-tree.h"

/*
  The kd-tree element pool functions handle the data structures
  (linear kd-tree, element pool, pair pool) holding all elements
  and sub-elements information.
  
  These functions replace those in "array.h".
*/

/*
  Design alternatives:
  
  - replace all indexes by direct pointers (-> more memory but less cpu)
  - remove builtin bounding-sphere culling (-> more pairs)
*/

// *********************************************************************
// Data structures
// *********************************************************************
#define KDTREE_INDEX	uint32_t

typedef struct kdtree_element_t kdtree_element;
typedef struct kdtree_pair_t kdtree_pair;
typedef struct kdtree_pair_buffer_t kdtree_pair_buffer;
typedef struct kdtree_entry_t kdtree_entry;
typedef struct kdtree_pool_t kdtree_pool;

struct kdtree_element_t {
	void *pointer;
	kdtree_sphere sphere;
	kdtree_pair *first;
};

struct kdtree_pair_t {
	KDTREE_INDEX idx_A, idx_B;
	kdtree_pair *next_A, *next_B;
};

struct kdtree_pair_buffer_t {
	size_t npairs;
	kdtree_pair_buffer *next;
	kdtree_pair pairs[];
};

struct kdtree_entry_t {
	kdtree_node path;
	KDTREE_INDEX idx; /* or a (kdtree_element *) */
};

struct kdtree_pool_t {
	KDTREE_INDEX nel, nel_alloc, nent;

	kdtree_element *elements;
	kdtree_entry *entries;
	
	size_t pair_gran;
	kdtree_pair_buffer *pair_buf;
	
	KDTREE_INDEX stats_out_of_bounds;
	KDTREE_INDEX stats_tree_collisions;
	KDTREE_INDEX stats_sphere_collisions;
	KDTREE_INDEX stats_unique_collisions;
};


// *********************************************************************
// Functions
// *********************************************************************
kdtree_pool *kdtree_pool_new(KDTREE_INDEX nelements, KDTREE_INDEX pairs_base);
void kdtree_pool_free(kdtree_pool *b);


// Offline linear kd-tree
int  kdtree_pool_add_sphere(kdtree_pool *b, kdtree_sphere *bs, void *obj);
void kdtree_pool_build_tree(kdtree_pool *b, kdtree_space *s);
void kdtree_pool_sort_tree(kdtree_pool *b);
int  kdtree_pool_sweep_tree(kdtree_pool *b);




#endif

