/*
    main.c - a visualization tool for simulation benchmarks
    
    Copyright (C) 2009  Laurent Poirrier
    This file is part of the lin-kd-tree library.

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/times.h>
#include <GL/gl.h>
#include <math.h>
#include "x_glx.h"
#include "jpeg_dev.h"
#include "vec_math.h"
#include "gen_graph.h"
#include "../genbench.h"
#include "main.h"

static float a = 0.0, b = 0.0, dist = -2.0;

static graph_surface *sphere;
static vec3 *sphere_normals;

#define HELP_MSG	\
"Usage: replay [-d dist] [-c divs] [-flat] [-nolight]\n" \
"              [-noblend] [-wireframe] [-mjpeg file] <file>\n"

#define R_FLATSHADING	1
#define R_NOLIGHT	2
#define R_NOBLEND	4
#define R_WIREFRAME	8
#define R_DEFAULT	0

// ******************************************************************
// Drawing basics
// ******************************************************************
static void gl_sphere(scalar radius)
{
	vec3 lpos;
	int f, v, vn;
	
	glBegin(GL_TRIANGLES);
	for (f = 0; f < sphere->nfaces; f++) {
		for (v = 0; v < 3; v++) {
			vn = sphere->faces[f].v[v];
			vec3_set_scaled(&lpos,
				&sphere->vertices[vn], radius);

			VEC_GL_V(glNormal3) (sphere_normals[vn].d);
			VEC_GL_V(glVertex3) (lpos.d);
		}
	}
	glEnd();

}

static void gl_setup(int rflags)
{
	float q[4];

	glDrawBuffer(GL_BACK);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	
	if (rflags & R_NOBLEND) {
		glDisable(GL_BLEND);
	} else {
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
	
	if (rflags & R_FLATSHADING) {
		glShadeModel(GL_FLAT);
	} else {
		glShadeModel(GL_SMOOTH);
	}
	
	
	glClearColor(.0,.0,.3,.0);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-.1,.1,-.075,.075,.10,6000.0);
	//glViewport(0, 0, w, h);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	if (rflags & R_NOLIGHT) {
		glDisable(GL_LIGHTING);
	} else {
		glEnable(GL_LIGHTING);
		q[0]=q[1]=q[2]=1.0; q[3]=.0;
		glLightfv(GL_LIGHT0, GL_POSITION, q);
		q[0]=q[1]=q[2]=q[3]=.0; glLightfv(GL_LIGHT0, GL_AMBIENT, q);
		q[0]=q[1]=q[2]=q[3]=.7; glLightfv(GL_LIGHT0, GL_DIFFUSE, q);
		q[0]=q[1]=q[2]=q[3]=1.0; glLightfv(GL_LIGHT0, GL_SPECULAR, q);
		glEnable(GL_LIGHT0);

		q[0]=q[1]=q[2]=q[3]=1.0;
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, q);
		q[0] = 50.0;
		glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, q);
		glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);
	}
	
	glColor4f(.0, .0, .0, 1.0);
	/*glEnable(GL_AUTO_NORMAL);
	glEnable(GL_NORMALIZE);*/

	if (rflags & R_WIREFRAME) {
		glPolygonMode(GL_FRONT, GL_LINE);
		glPolygonMode(GL_BACK, GL_LINE);
	} else {
		glPolygonMode(GL_FRONT, GL_FILL);
		glPolygonMode(GL_BACK, GL_FILL);
	}
}


void draw_all(x_display *d, x_window *w, char *mjpeg_file)
{
	int i, width, height;
	double ar, ag, ab, ax, ay, az;
	
	x_get_window_size(w, &width, &height);
	glViewport(0, 0, width, height);

	//glClearColor(.12,.12,.12,.0);
	glClearColor(1.0,1.0,1.0,.0);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	glTranslatef(.0, .0, dist);
	glRotatef(b, 1.0, .0, .0);
	glRotatef(a, .0, 1.0, .0);

	//a -= 100.0 * delta_t;
	//b += 0.01; // 10.0 * delta_t;

	for (i = 0; i < gb_nobj(); i++) {
		gb_obj_coord(i, &ax, &ay, &az);
		gb_obj_color(i, &ar, &ag, &ab);
		glColor4d(ar, ag, ab, 1.0);

		glPushMatrix();
		glTranslated(ax, ay, az);
		gl_sphere(gb_obj_radius(i));
		glPopMatrix();
	}
	
	//phys_draw_checkerboard(0.50);
	
	x_swap_buffers(w);
	//x_flush(d);
	x_sync(d);

	if (mjpeg_file) { // This is quite inefficient but works for now
		void *pix;

		pix = malloc(width * height * 3);
		if (pix) {
			//glPixelStore(); glPixelTransfer(); glPixelMap();
			glPixelStorei(GL_PACK_ALIGNMENT, 1);
			glReadPixels(0, 0, width, height, GL_RGB,
				GL_UNSIGNED_BYTE, pix);
			jpegdev_compress_file(mjpeg_file, pix, width, height,
				JPEGDEV_PACKED_RGB|JPEGDEV_BOTTOM_UP
				|JPEGDEV_APPEND_FILE);

			free(pix);
		}
	}
}

int main(int argc, char **argv, char **envp)
{
	int fr, complexity = 2, rflags = R_DEFAULT, dump_active = 0;
	int t, k, wk, nextframe, continuous = 0;
	char *fname = NULL, *mjpeg_file = NULL;
	x_display *d;
	x_window *w;

	for (t = 1; t < argc; t++) {
		if ((!strcmp(argv[t], "-h")) || (!strcmp(argv[t], "--help"))) {
			printf(HELP_MSG);
			return(0);
		} else if ((!strcmp(argv[t], "-mjpeg")) && (t + 1 < argc)) {
			mjpeg_file = argv[++t];
		} else if ((!strcmp(argv[t], "-c")) && (t + 1 < argc)) {
			complexity = strtol(argv[++t], NULL, 0);
		} else if ((!strcmp(argv[t], "-d")) && (t + 1 < argc)) {
			dist = -strtod(argv[++t], NULL);
		} else if (!strcmp(argv[t], "-flat")) {
			rflags |= R_FLATSHADING;
		} else if (!strcmp(argv[t], "-nolight")) {
			rflags |= R_NOLIGHT;
		} else if (!strcmp(argv[t], "-noblend")) {
			rflags |= R_NOBLEND;
		} else if (!strcmp(argv[t], "-wireframe")) {
			rflags |= R_WIREFRAME;
		} else if (fname) {
			printf(HELP_MSG);
			return(1);
		} else {
			fname = argv[t];
		}
	}
	
	if (!fname) {
		printf(HELP_MSG);
		return(0);
	}

	if (!strcmp(fname, "-")) {
		if (gb_load_sim_file(stdin))
			return(-1);
	} else {
		if (gb_load_sim(fname))
			return(1);
	}
	
	d = x_open_display();
	if (!d) {
		printf("Could not open display\n");
		return(1);
	}
	w = x_create_window(d, 0, 0, 800, 600, 0);
	if (!w) {
		printf("Can't create OpenGL window\n");
		return(1);
	}

	sphere = graph_gen_sphere(complexity);
	sphere_normals = graph_vertex_normals(sphere);
	
	gl_setup(rflags);
	
	continuous = 0;
	fr = -1;
	do {
		wk = !continuous;
		nextframe = 0;
		do {
			k = x_get_input(d, wk);
			wk = 0;

			if (k == 36) {
				continuous = !continuous;
			} else if ((k == 100) || (k == 113)) {
				a += 10.0;
			} else if ((k == 102) || (k == 114)) {
				a -= 10.0;
			} else if ((k == 99) || (k == 112)) {
				b += 10.0;
			} else if ((k == 105) || (k == 117)) {
				b -= 10.0;
			} else if ((k == 98) || (k == 111)) {
				dist += 1.0;
			} else if ((k == 104) || (k == 116)) {
				dist -= 1.0;
			} else if (k == 9) {
			} else if (k == 22) {
				dump_active = !dump_active;
				if (mjpeg_file) {
					printf("********************** mjpeg "
						"output: %d\n", dump_active);
				}
			} else if (k == 65) {
				if (continuous) {
					continuous = 0;
				} else {
					nextframe = 1;
				}
			} else if ((k == 41) || (k == 46) || (k == 56)
			|| (k == 52)) {
				if (k == 41) rflags ^= R_FLATSHADING;
				else if (k == 46) rflags ^= R_NOLIGHT;
				else if (k == 56) rflags ^= R_NOBLEND;
				else if (k == 52) rflags ^= R_WIREFRAME;
				
				gl_setup(rflags);
			} else if ((k == 86) || (k == 61)
			|| (k == 21) || (k == 82)) {
				if ((k == 86) || (k == 61)) complexity++;
				else complexity--;
				
				if (complexity < 0) complexity = 0;
				printf("sphere complexity: %d\n", complexity);
				free(sphere);
				free(sphere_normals);
				sphere = graph_gen_sphere(complexity);
				sphere_normals = graph_vertex_normals(sphere);
			} else if (k == 82) {
			} else if (k != 0) {
				printf("x_get_input(): %d\n", k);
			}
		} while ((k != 0) && (k != 9));
		
		if ((continuous) || (nextframe)) {
			if (!gb_load_frame()) fr++;
		}
		
		draw_all(d, w, dump_active ? mjpeg_file : NULL);
		printf("frame %d\n", fr);
	} while(k != 9);

	x_destroy_window(w);
	x_close_display(d);

	return(0);
}


